'use strict';

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.tabulate = tabulate;
exports.consoleTable = consoleTable;
exports.markdownTable = markdownTable;

var _cliTable = require('cli-table');

var _cliTable2 = _interopRequireDefault(_cliTable);

var _markdownTable = require('markdown-table');

var _markdownTable2 = _interopRequireDefault(_markdownTable);

var _colors = require('colors');

var _colors2 = _interopRequireDefault(_colors);

var _lodash = require('lodash');

var _lodash2 = _interopRequireDefault(_lodash);

var _jsonpath = require('jsonpath');

var _jsonpath2 = _interopRequireDefault(_jsonpath);

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

/*
This module contains functionality for outputting metric data in tabular form.

Exposed Functionality:

tabulate function - Generates an array of arrays representing a table from a JSON data
object, configurable using a tableOptions object.

consoleTable function - Outputs a string representation of tabulated data, formatted
for console display.

markdownTable function - Outputs a string representation of tabulated data in Markdown format.
*/

function tabulate({ dataObject, tableOptions, shouldStyle = false }) {
  if (!_lodash2.default.isObject(dataObject)) throw new Error('A data object must be provided');
  if (!_lodash2.default.isArray(tableOptions.tableHeaders)) throw new Error('Table options must include a headers array');
  if (!_lodash2.default.isArray(tableOptions.columns)) throw new Error('Table options must include a columns array');
  if (tableOptions.tableHeaders.length !== tableOptions.columns.length) {
    throw new Error('The number of headers must match the number of columns');
  }

  const HIDE_ROW_INDICATOR = 'HIDE';
  const MISSING_VALUE_INDICATOR = shouldStyle ? _colors2.default.red('NO VALUE') : 'NO VALUE';

  const tableHeaders = tableOptions.tableHeaders.map(h => shouldStyle ? _colors2.default.bold(h) : h);

  const getFirstMatch = (obj, path) => {
    let fullPath;
    if (path.startsWith('$')) fullPath = path;else if (path.startsWith('.')) fullPath = `$${path}`;else fullPath = `$.${path}`;

    return _jsonpath2.default.query(obj, fullPath)[0];
  };

  // Get the collection of data from which the table will be constructed
  const collection = tableOptions.collection ? getFirstMatch(dataObject, tableOptions.collection) : dataObject;

  const getTableRow = collectionElement => {
    const getSingleValue = accessor => {
      if (_lodash2.default.isObject(accessor)) {
        if (!('datasets' in accessor) || !(_lodash2.default.isString(accessor.datasets) || _lodash2.default.isArray(accessor.datasets))) {
          throw new Error('Dataset(s) must be provided as a path string or an array');
        }
        if (_lodash2.default.isArray(accessor.datasets) && !('valueTransform' in accessor || _lodash2.default.isFunction(accessor.valueTransform))) {
          throw new Error('A value transform must be provided when an array of datasets is given');
        }

        let value;
        if (_lodash2.default.isString(accessor.datasets)) {
          value = getFirstMatch(collectionElement, accessor.datasets);
          value = 'valueTransform' in accessor ? accessor.valueTransform(value) : value;
        } else {
          const datasetRow = accessor.datasets.map(path => getFirstMatch(collectionElement, path));
          value = accessor.valueTransform(...datasetRow);
        }

        return !value && accessor.hideRowIfFalsy ? HIDE_ROW_INDICATOR : value;
      } else if (_lodash2.default.isString(accessor)) {
        return getFirstMatch(collectionElement, accessor);
      }
      throw new Error('Columns must be objects or strings');
    };

    const valueAccessors = tableOptions.columns;
    const tableRow = valueAccessors.map(getSingleValue);

    return tableRow;
  };

  let tableRows = [];
  if (_lodash2.default.isArray(collection)) {
    tableRows = collection.map(getTableRow);
  } else if (_lodash2.default.isObject(collection)) {
    tableRows = Object.keys(collection).map(key => getTableRow(collection[key]));
  } else {
    throw new Error('The collection must be an array or an object');
  }

  tableRows = tableRows.filter(row => !row.some(value => value === HIDE_ROW_INDICATOR)).map(row => row.map(value => value || MISSING_VALUE_INDICATOR));

  return _lodash2.default.isEmpty(tableRows) ? [] : [tableHeaders].concat(tableRows);
}

function consoleTable({ dataObject, tableOptions }) {
  const tableData = tabulate({ shouldStyle: true, tableOptions, dataObject });
  if (_lodash2.default.isEmpty(tableData)) return '';

  const prettyTable = new _cliTable2.default();
  prettyTable.push(...tableData);

  return prettyTable.toString();
}

function markdownTable({ dataObject, tableOptions }) {
  const tableData = tabulate({ shouldStyle: false, tableOptions, dataObject });
  if (_lodash2.default.isEmpty(tableData)) return '';

  return (0, _markdownTable2.default)(tableData);
}